#//#
# FuseVFS is a Tcl extension that exports Tcl's VFS on userspace for
# all system processes to access. The Tcl VFS is exported by means of
# the FUSE kernel subsystem. This extension relies on tcl-fuse.
#
# The following FUSE operations have been implemented and tested:
#     getattr
#     readdir
#     readlink
#     open
#     read
#     write
#     rename
#     mknod
#     mkdir
#     unlink
#     rmdir
#     chmod
#     chown
#     symlink
#     link
#     truncate
#     access
#     utimens
#     truncate
#     access
#     flush
#     fsync
# 
# Whereas the following were not implemented:
#     statfs
#     setxattr
#     getxattr
#     listxattr
#     removexattr
#     opendir
#     releasedir
#     fsyncdir
#     create
#     ftruncate
#     fgetattr
#     destroy
#     lock
#     bmap
#     release
# either because they weren't necessary (ftruncate, fgetattr, fsyncdir),
# applicable (bmap), or the necessary information are not provided by Tcl
# at all (statfs, setxattr, getxattr, listxattr, removexattr). For
# example, Tcl's VFS does not support "exteded attributes", hence we
# cannot provided them to FUSE.
# 
# An other limitation of Tcl's VFS is that it does not support create of
# fifos, sockets, & device files, and hence it is not supported in FuseVFS
# either.
# 
# The release, opendir, and releasedir operations are not necessary due to
# the state-less nature of this implementation. A state-full version is
# planned for implementation, which will have greater performance and more
# thoroughly test tcl-fuse.
#
# USAGE:
#   package require fusevfs
#   fusevfs <root> <mntpoint>
#
#   <root>: The Tcl VFS directory to export.
#   <mntpoint>: The real directory to mount the <root> VFS directory 
#               above. 
#
# In the procedure definitions of this file, the "fileinfo" argument
# is repeatedly used, which is a dictionary with the following keys:
#       flags : Open flags for the file, as those supported by open(n).
#       writepage   : In case of a write operation indicates if this was 
#                     caused by a writepage.
#       direct_io   : Can be filled in by open, to use direct I/O on this 
#                     file.
#       keep_cache  : Can be filled in by open, to indicate, that cached 
#                     file data need not be invalidated.
#       flush       : Indicates a flush operation.  Set in flush operation, 
#                     also maybe set in highlevel lock operation and 
#                     lowlevel release operation.
#       file_handle : The file handle that is associated with the file. In
#                     state-less operation this value is undefined (can be
#                     anything).
#       lock_owner  : Lock owner id.  Available in locking operations and 
#                     flush.
#
# TODO: Consult umask when creating new files and directories.
# TODO: Enforce individual user permissions and authorization.
# TODO: Create a state-full version of this file.
#
#//#

package require fuse
package provide fusevfs 1.0

namespace eval fusevfs {
namespace export mount unmount

# Export a Tcl VFS directory hierarchy, by mounting it on a real 
# system directory.
#
# @param root The Tcl VFS directory to export.
# @param mntpoint The real system directory to mount to.
# @return Nothing
proc mount {root mntpoint} {
    fuse create ::fusevfs::fusevfs \
            -userdata $root \
            -getattr ::fusevfs::Getattr \
            -readdir ::fusevfs::Readdir \
            -readlink ::fusevfs::Readlink \
            -open ::fusevfs::Open \
            -read ::fusevfs::Read \
            -write ::fusevfs::Write \
            -rename ::fusevfs::Rename \
            -mknod ::fusevfs::Mknod \
            -mkdir ::fusevfs::Mkdir \
            -unlink ::fusevfs::Unlink \
            -rmdir ::fusevfs::Rmdir \
            -chmod ::fusevfs::Chmod \
            -chown ::fusevfs::Chown \
            -symlink ::fusevfs::Symlink \
            -link ::fusevfs::Link \
            -truncate ::fusevfs::Truncate \
            -access ::fusevfs::Access \
            -utimens ::fusevfs::Utimens
    #::fusevfs::fusevfs $mntpoint -ononempty ;# -oallow_other
    ::fusevfs::fusevfs $mntpoint -f -d -ononempty ;# -oallow_other
}

# Unmount a previously mounted FuseVFS filesystem.
#
# @param mntpoint The mount point where the filesystem is mounted.
proc unmount {mntpoint} {
    fuse unmount $mntpoint
}

# Retrieve file and directory information, similar to lstat(2).
#
# The "st_dev" and "st_blksize" fields are
# ignored.  The "st_ino" field is ignored except if the "use_ino"
# mount option is given.
#
# The returned information
# include everything that "file stat" reports, and in the same
# format. Additionally, it supports the following dictionary keys
# that are not present in "file stat":
#   rdev : The device ID, if the file is a device special file.
#   blksize : The blocksize for filesystem I/O.
#   blocks : The current number of blocks allocated for the file.
#
# If a key attribute is not present, the default is applied which
# is zero in all cases.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @return A Tcl dictionary like the serialized array that 
# "file stat" generates.
# @error See lstat(2).
proc Getattr {root context path} {
    file lstat [file join $root $path] result
    return [array get result]
}

# Open a file, similar to open(2).
#
# No creation, or truncation flags (CREAT, EXCL, TRUNC)
# will be passed to this procedure via the "flags"
# dictionary key of "fileinfo", as well as the NOCTTY flag.
#
# Open should check if the operation
# is permitted for the given flags. Optionally open may also
# return an arbitrary filehandle (integer) that will be passed
# to all file operations that receive the "fileinfo" dictionary.
#
# The filesystem may also implement stateless file I/O and not
# return anything (empty list).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path, including the 
# open flags.
# @return 
# Nothing for stateless operation, or an integer to associate 
# with the open file. This number (or file descriptor/handle) 
# will be passed-on to every operation that follows, by
# means of the "file_handle" dictionary key in the "fileinfo"
# dictionary.
# @error See open(2).
proc Open {root context path fileinfo} {
    set target [file join $root $path]

    # IMPORTANT NOTE:
    # If the file does not exist it is created, and the default 
    # permissions are 666. "umask" value is not considered.

    # "open" checks permissions.
    close [open $target [dict get $fileinfo flags]]
    file attributes $target -owner [dict get $context uid] -group [dict get $context gid]
    return ;# if a number is returned, then it becomes the associated file descriptor.
}

# Read the contents of a directory, similar to readdir(3).
# 
# You must check that the target is in fact a directory.
# It is also legal for the target to be a symling that
# itself points to a directory.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @return A list of name for all the contents of the directory.
# @error See readdir(2).
proc Readdir {root context path fileinfo} {
    set target [file join $root $path]

    # NOTE:
    # "file isdirectory" returns true also for
    # symlinks that point to a directory.
    if {! [file isdirectory $target]} {
        return -code error -errorcode [list POSIX ENOTDIR {}]
    }

    return [glob -tails -nocomplain -directory $target *]
}

# Read from a file, similar to read(2).
#
# You must check that the target is in fact a file.
# EOF is signified by returning less than the number
# of bytes requested, or the empty list.
#
# An exception to this is when the
# "direct_io" mount option is specified, in which case the return
# value of the read system call will reflect the return value of
# this operation.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @param size Number of bytes to read.
# @param offset File offset (in bytes) to start reading from.
# @return The data read or the empty list. Can be binary.
# @error See read(2).
proc Read {root context path fileinfo size offset} {
    set target [file join $root $path]

    set len [file size $target]
    if {$offset < $len} {
        if {$offset + $size > $len} {
            set size [expr {$len - $offset}]
        }
        set ch [open $target "r"]
        fconfigure $ch -blocking no -buffering none -translation binary
        seek $ch $offset
        set data [read $ch $size]
        close $ch
        return $data
    } 
    
    return
}

# Change the name and/or location of a file, similar to rename(2).
#
# You must check that the target is in fact a file.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param newpath The new location and name for the entity.
# @return The return value is ignored.
# @error See rename(2).
proc Rename {root context path newpath} {
    file rename -force [file join $root $path] [file join $root $newpath]
}

# Create a new special or ordinary file, similar to mknod(2).
#
# This is called for creation of all non-directory, non-symlink
# nodes.  If the FUSE filesystem defines a create operations, 
# then for regular files that will be called instead.
#
# You must check that the target does not already exist.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param type The node type to create. One of the following:
# "file", "characterSpecial", "blockSpecial", "fifo", "socket".
# @param perm A number describing the permissions for 
# the file, such as 0777 but in decimal.
# @param dev If the file type is "characterSpecial" or 
# "blockSpecial" then dev specifies the major and minor numbers 
# (in a list) of the newly created device special file; 
# otherwise it is set to the empty list.
# @return The return value is ignored.
# @error See mknod(2).
proc Mknod {root context path type perm dev} {
    set target [file join $root $path]
    
    if {[file exists $target]} {
        return -code error -errorcode [list POSIX EEXIST {}]
    }

    if {$type ne "file"} {
        # ATTENTION: 
        #   TclVFS does not support creation of device files,
        #   FIFOs or sockets. Therefore we only accept 
        #   creation of regular files.
        return -code error -errorcode [list POSIX EINVAL {}]
    }

    close [open $target {RDONLY CREAT}]
    file attributes $target -permissions $perm
    file attributes $target -owner [dict get $context uid] -group [dict get $context gid]
}

# Create a new directory, similar to mkdir(2).
#
# You must check that the target does not already exist
# (not necessarily as directory).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param perm A number describing the permissions for 
# the file, such as 0777 but in decimal.
# @return The return value is ignored.
# @error See mkdir(2).
proc Mkdir {root context path perm} {
    set target [file join $root $path]
    
    if {[file exists $target]} {
        return -code error -errorcode [list POSIX EEXIST {}]
    }

    file mkdir $target
    file attributes $target -permissions $perm
    file attributes $target -owner [dict get $context uid] -group [dict get $context gid]
}

# Delete a name and possibly the file it refers to from
# the filesystem. It is similar to unlink(2).
#
# The name can refer to a file, symlink, fifo, socket or device.
# 
# You must check that the target is not a directory.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @return The return value is ignored.
# @error See unlink(2).
proc Unlink {root context path} {
    set target [file join $root $path]

    if {[file type $target] eq "directory"} {
        return -code error -errorcode [list POSIX EISDIR {}]
    }

    # If $target is symlink, the following will remove
    # the symlink, not its target.
    file delete -- $target
}

# Delete a directory, similar to rmdir(2).
#
# You must check that the target is a directory.
# The directory must be empty.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @return The return value is ignored.
# @error See rmdir(2).
proc Rmdir {root context path} {
    set target [file join $root $path]
    
    if {[file type $target] ne "directory"} {
        return -code error -errorcode [list POSIX ENOTDIR {}]
    }

    
    if {[glob -nocomplain -directory $target *] ne {}} {
        return -code error -errorcode [list POSIX ENOTEMPTY {}]
    }

    file delete -- $target
}

# Change permissions of a file, directory or other 
# filesystem entity, similar to chmod(2).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param perm A number describing the permissions for 
# the file, such as 0777 but in decimal.
# @return The return value is ignored.
# @error See chmod(2).
proc Chmod {root context path perm} {
    file attributes [file join $root $path] -permissions $perm
}

# Read the target of a symbolic link, similar to readlink(2).
#
# You must check that target is in fact a symlink.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param perm A number describing the permissions for 
# the file, such as 0777 but in decimal.
# @return The return value is ignored.
# @error See readlink(2).
proc Readlink {root context path} {
    # This will generate a POSIX error if target is not a link.
    return [file readlink [file join $root $path]]
}

# Change ownership for a file, directory or other 
# filesystem entity, similar to chown(2).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param owner The new owner of the file.
# @param group The new group for the file.
# @return The return value is ignored.
# @error See chown(2).
proc Chown {root context path owner group} {
    file attributes [file join $root $path] -owner $owner -group $group
}

# Create a symbolic link to a filesystem entity, similar to symlink(2).
#
# You must check that a filesystem entity does not already exist
# in the place where the symlink is requested to be created.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param link The location of the symlink, relatively to $path.
# @return The return value is ignored.
# @error See symlink(2).
proc Symlink {root context path link} {
    # WARNING
    # symlink(2) allows creation of symlinks to non-existing targets.
    # These symlinks are called dangling links. Unfortunately, Tcl's
    # "file link" command cannot create such symlinks and therefore 
    # the TclFUSE implementation is not 100% compatible to symlink(2).
    # This can be confusing to users, because standard commands like
    # "ln" will behave differently under TclFUSE compared to other
    # filesystesm.
    # Again, this is a limitation of Tcl's "file link" command.
    # The error message:
    #   -code 1 -level 0 -errorcode NONE -errorinfo {could not create new link "/tmp/fuse/softlink": target "fuse/a" doesn't exist
    #       while executing
    #   "file link /tmp/fuse/softlink fuse/a"} -errorline 1
    #
    set symlink [file join $root $link]

    if {[file exists $symlink]} {
        return -code error -errorcode [list POSIX EEXIST {}]
    }
    file link -symbolic $symlink $path
}

# Create a new hard link for a filesystem entity, similar to link(2).
#
# You must check that a filesystem entity does not already exist
# in the place where the symlink is requested to be created.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param link The location of the hardlink, relatively to $path.
# @return The return value is ignored.
# @error See link(2).
proc Link {root context path link} {
    set target [file join $root $path]
    set hardlink [file join $root $link]

    if {[file exists $hardlink]} {
        return -code error -errorcode [list POSIX EEXIST {}]
    }
    file link -hard $hardlink $target
}

# Truncate a file to a specified length, similar to truncate(2).
#
# You must check that target is not a directory.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param size The new size of the file after truncation.
# @return The return value is ignored.
# @error See truncate(2).
proc Truncate {root context path size} {
    set target [file join $root $path]

    if {[file isdirectory $target]} {
        return -code error -errorcode [list POSIX EISDIR {}]
    }
    set fd [open $target "a"]
    chan truncate $fd $size
    close $fd
}

# Checks whether the process requesting the operation can access 
# the file pathname. 
#
# This will be called for the access() system call.  If the
# "default_permissions" mount option is given, this method is not
# called.
#
# If pathname is a symbolic link, it is 
# dereferenced first. This operation is similar to access(2).
#
# The mode specifies the accessibility check(s) to be performed, 
# and is either the value "exists", or a list consisting a
# combination of the words "read", "write", and "execute". "exists" 
# tests for the existence of the file. The other three test for
# read, write, and execute permissions, respectively.
#
# Unlike in access(2), the check is done using the calling process's
# effective UID/GUI. This can create some inconsistency while using
# the TclVFS in some usage scenarions.
#
# If the calling process is privileged (i.e., its real UID is zero), 
# then an "exists" check is successful for a regular file if "execute"
# permission is enabled for any of the file owner, group, or other.
#
# Note: Generate the EACCES error when the requested privileges are
# not met. In case of multiple flags in $mask, all of them must be
# met.
#
# This method is not called under Linux kernel versions 2.4.x
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param mask Either "exists" or a list of "read", "write", "execute".
# @return The return value is ignored.
# @error See access(2).
proc Access {root context path mask} {
    set target [file join $root $path]

    if {[file type $target] eq "link"} {
        set target [file readlink $target]
    }
    
    if {$mask eq "exists"} {
        if {! [file exists $target]} {
            return -code error -errorcode [list POSIX EACCES {}]
        }
    } 

    if {"read" in $mask} {
        if {! [file readable $target]} {
            return -code error -errorcode [list POSIX EACCES {}]
        }
    } 

    if {"write" in $mask} {
        if {! [file writable $target]} {
            return -code error -errorcode [list POSIX EACCES {}]
        }
    } 

    if {"execute" in $mask} {
        if {! [file executable $target]} {
            return -code error -errorcode [list POSIX EACCES {}]
        }
    } 
}

# Change the access and modification times of a file with
# second resolution.
#
# Utimens can be performed on a directory, in addition to other
# filesystem entities.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @param atime Last access time in number of seconds since 1 
# January 1970.
# @param mtime Last modification time in number of seconds since
# 1 January 1970.
# @return The return value is ignored.
# @error All POSIX codes.
proc Utimens {root context path atime mtime} {
    set target [file join $root $path]
    file atime $target $atime
    file mtime $target $mtime
}

# Synchronize file contents, similar to write(2).
#
# Write should either write the exact number of bytes requested
# or generate an error. No partial writes should happen.
# An exception to this is when the "direct_io"
# mount option is specified (see read operation).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @param data The data to be written. Can be binary.
# @param offset The offset in the file to write the data.
# @return The number of bytes written.
# @error See write(2).
proc Write {root context path fileinfo data offset} {
    set target [file join $root $path]
    set ch [open $target "a"]
    fconfigure $ch -blocking no -buffering none -translation binary
    seek $ch $offset
    puts $ch $data
    close $ch
    return [string length $data]
}

######################################################################
# Due to the stateless nature of this implementation, we don't
# need to implement the following procedures:
#    release, flush, fsync, opendir, releasedir, fsyncdir, 
#    ftruncate, fgetattr, and destroy
# Furthermore, we don't have to implement the create procedure
# because we use mknod instead.
######################################################################

# Release an open file.
#
# Release is called when there are no more references to an open
# file: all file descriptors are closed and all memory mappings
# are unmapped.
#
# For every open() call there will be exactly one release() call
# with the same flags and file descriptor.  It is possible to
# have a file opened more than once, in which case only the last
# release will mean, that no more reads/writes will happen on the
# file.
#
# After a successfull execution of this procedure the fileinfo's
# file_handle key is reset to 0, as we assume that any file handle
# has been closed.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @return The return value is ignored.
# @error All POSIX codes.
proc Release {root context path fileinfo} {
    # We are state-less, we don't need to clean-up anything here.
}

# Possibly flush cached data.
#
# ATTENTION: This is not equivalent to fsync(2).  It's not a
# request to sync dirty data.
#
# Flush is called on each close() of a file descriptor. So if a
# filesystem wants to return write errors in close() and the file
# has cached dirty data, this is a good place to write back data
# and return any errors. Since many applications ignore close()
# errors this is not always useful.
#
# NOTE: The flush procedure may be called more than once for each
# open. This happens if more than one file descriptor refers
# to an opened file due to dup(), dup2() or fork() calls.  It is
# not possible to determine if a flush is final, so each flush
# should be treated equally. Multiple write-flush sequences are
# relatively rare, so this shouldn't be a problem.
#
# Filesystems shouldn't assume that flush will always be called
# after some writes, or that if will be called at all
#
# You must check that target is a file (not a directory).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @return The return value is ignored.
# @error All POSIX codes.
proc Flush {root context path fileinfo} {
    # We are state-less, we don't anything to flush here.
}

# Synchronize file contents, similar to fsync(2).
#
# fsync can be performed on a directory, in addition to other
# filesystem entities.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @param datasync Boolean. If set then only data should be 
# synced, not metadata.
# @return The return value is ignored.
# @error See fsync(2).
proc Fsync {root context path fileinfo datasync} {
    # We are state-less, we don't anything to sync here.
}

# Open directory, similar to opendir(3).
#
# This method should check if the open operation is permitted for
# the directory.
#
# This procedure can return an arbitrary file handle (integer)
# that will be present in fileinfo in other all other directory
# stream operations (readdir, releasedir, fsyncdir).
#
# The filesystem may also implement stateless directory I/O and not
# return anything (empty list), though that makes it impossible to
# implement standard conforming directory stream operations in
# case the contents of the directory can change between opendir
# and releasedir.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @return 
# Nothing for stateless operation, or an integer to associate 
# with the open directory. This number (or file descriptor/handle) 
# will be passed-on to readdir and release dir operations that will
# follow, by means of the "file_handle" dictionary key in the "fileinfo"
# dictionary.
# @error See opendir(3).
proc Opendir {root context path fileinfo} {
    # Since we are state-less, we don't open/close the dir first before
    # we read from it, hence we only have the readdir procedure.
}

# Release an opened directory, similar to closedir(3).
#
# This method should check if the open operation is permitted for
# the directory.
#
# After a successfull execution of this procedure the fileinfo's
# file_handle key is reset to 0, as we assume that any file handle
# has been closed.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @return The return value is ignored.
# @error See opendir(3).
proc Releasedir {root context path fileinfo} {
    # Since we are state-less, we don't open/close the dir first before
    # we read from it, hence we only have the readdir procedure.
}

# Synchronize directory contents.
#
# If the datasync parameter is non-zero, then only the user data
# should be flushed, not the meta data
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @param datasync Boolean. If set then only data should be 
# synced, not metadata.
# @return The return value is ignored.
# @error All POSIX codes.
proc Fsyncdir {root context path fileinfo datasync} {
    # We are state-less, we don't anything to sync here.
}

# Create and open a file.
#
# If the file does not exist, first create it with the specified
# mode, and then open it.
#
# No NOCTTY flag will be passed to this procedure via the "flags"
# dictionary key of "fileinfo"..
#
# If this method is not implemented or under Linux kernel
# versions earlier than 2.6.15, the mknod and open procedures 
# will be called instead.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path, including the 
# open flags.
# @param perm A number describing the permissions for 
# the file, such as 0777 but in decimal.
# @return 
# Nothing for stateless operation, or an integer to associate 
# with the open file. This number (or file descriptor/handle) 
# will be passed-on to every operation that follows, by
# means of the "file_handle" dictionary key in the "fileinfo"
# dictionary.
# @error All POSIX codes.
proc Create {root context path fileinfo perm} {
    # We use mknod instead.
}

# Truncate an open file to a specified length, similar to ftruncate(2).
#
# You must check that target is not a directory.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @param size The new size of the file after truncation.
# @return The return value is ignored.
# @error See ftruncate(2).
proc Ftruncate {root context path fileinfo size} {
    # We use truncate instead.
}

# Get attributes from an open file, similar to fstat(2).
#
# This procedure is called instead of the getattr procedure if the
# file information is available.
#
# Currently this is only called after the create procedure if that
# is implemented (see above). Later it may be called for
# invocations of fstat too
# Retrieve file and directory information, similar to lstat(2).
#
# The "st_dev" and "st_blksize" fields are
# ignored.  The "st_ino" field is ignored except if the "use_ino"
# mount option is given.
#
# The returned information
# include everything that "file stat" reports, and in the same
# format. Additionally, it supports the following dictionary keys
# that are not present in "file stat":
#   rdev : The device ID, if the file is a device special file.
#   blksize : The blocksize for filesystem I/O.
#   blocks : The current number of blocks allocated for the file.
#
# If a key attribute is not present, the default is applied which
# is zero in all cases.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @return A Tcl dictionary like the serialized array that 
# "file stat" generates.
# @error See fstat(2).
proc Fgetattr {root context path fileinfo} {
    # We use getattr instead.
}

# Clean up filesystem.
#
# Called on filesystem exit.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @return The return value is ignored.
# @error All POSIX codes.
proc Destroy {root context} {
    # We are state-less, we don't need to clean-up anything here.
}

###############################################
# All the remaining are not applicable to Tcl VFS,
# for one reason or another.
###############################################

# Get file system statistics, similar to statvfs(2).
#
# The result is a dictionary consisting of the keys:
#   
# bsize    file system block size 
# frsize   fragment size 
# blocks   size of fs in f_frsize units 
# bfree    free blocks 
# bavail   free blocks for non-root 
# files    inodes
# ffree    free inodes 
# favail   free inodes for non-root 
# fsid     file system ID 
# flag     mount flags 
# namemax  maximum filename length 
#
# Currently the frsize, favail, fsid and flag keys are ignored.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @return A dictionary as described above.
# @error See statvfs(2).
proc Statfs {root context path} {
    # Tcl's VFS cannot provide us with the necessary information.
}

# Set an extended attribute, similar to setxattr(2).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param name The name of the extended attribute.
# @param value The new value for the attribute.
# @return The return value is ignored.
# @error See setxattr(2).
proc Setxattr {root context path name value} {
    # Tcl's VFS does not support extended attributes.
}

# Retrieve an extended attribute value, similar to getxattr(2).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param name The name of the extended attribute.
# @param size The maximum size in bytes that the returned value can have,
# excluding the null-terminating character if the result is a string.
# @return The value of the attribute. Can be binary.
# @error See getxattr(2).
proc Getxattr {root context path name size} {
    # Tcl's VFS does not support extended attributes.
}

# List extended attributes, similar to listxattr(2).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param size The maximum size in bytes that the returned value can have.
# In the available size calculation you should subtruct the number of 
# list items in the result, due to a null-terminating character for each one.
# @return A list of attribute name strings, or the empty list.
# @error See listxattr(2).
proc Listxattr {root context path size} {
    # Tcl's VFS does not support extended attributes.
}

# Remove extended attributes, similar to removexattr(2).
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param name The name of the extended attribute.
# @return The return value is ignored.
# @error See removexattr(2).
proc Removexattr {root context path name} {
    # Tcl's VFS does not support extended attributes.
}

# Map block index within file to block index within device.
#
# This makes sense only for block device backed filesystems
# mounted with the "blkdev" option.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param blocksize Device blocksize.
# @param index Block index.
# @return The return value is ignored.
# @error All POSIX codes.
proc Bmap {root context path blocksize index} {
    # Tcl's VFS is not necessarily backed by a block 
    # device, therefore not applicable.
}

# Perform POSIX file locking operation, as those performed
# with fcntl(2).
#
# See: http://www.delorie.com/gnu/docs/glibc/libc_264.html
#
# The cmd argument will be either "getlock", "setlock" or 
# "setlockblock". The meaning of each:
#
# getlock : To specify that it should get information about 
#           a lock.
# setlock : To specify that it should set or clear a lock.
# setlockblock : Like setlock, but causes the process to 
#                block (or wait) until the request can be 
#                specified.
#
# The argument lock is a dictionary with the following keys:
#   type
#       Specifies the type of the lock; one of "readlock", 
#       "writelock" or "unlock".
#   origin
#       Like the origin argument of seek(n). Accepts one of:
#       "start", "current", or "end".
#   offset
#       This specifies the offset of the start of the region 
#       to which the lock applies, and is given in bytes relative 
#       to the point specified by origin.
#   length
#       This specifies the length of the region to be locked. 
#       A value of 0 is treated specially; it means the region 
#       extends to the end of the file.
#   
# The origin key will always be set to "start".
#
# To check lock ownership, the lock_owner key of the fileinfo
# dictionary must be used, which holds the process ID of the
# process holding the lock. It is only significant when cmd
# argument is "getlock".
#
# For "getlock" operation, FUSE (on its own) will first check currently
# held locks, and if a conflicting lock is found it will return
# information without calling this procedure. This ensures, that
# for local locks the pid key is correctly filled in. The
# results may not be accurate in case of race conditions and in
# the presence of hard links, but it's unlikly that an
# application would rely on accurate getlock results in these
# cases. If a conflicting lock is not found, this method will be
# called, and the filesystem may return a non-empty number that
# will become the owning pid.
#
# The pid of the process performing the locking operation can be
# found from dictionary context.
#
# NOTE: if this method is not implemented, the kernel will still
# allow file locking to work locally. Hence it is only
# interesting for network filesystems and similar.
#
# @param root The VFS root that is exported via FUSE.
# @param context A list with the uid, gid and pid of the process
# performing the operation.
# @param path The file system entity to operate on.
# @param fileinfo File information for $path.
# @param cmd Either "getlock", "setlock" or "setlockblock".
# @param lock A dictionary. Read description above for its 
# constitution. 
# @return The pid number that holds the lock or nothing if the file
# region is not locked. This is when cmd is "getlock", in the other cases
# the return value is ignored.
# @error See fcntl(2).
proc Lock {root context path fileinfo cmd lock} {
    # Left unimplemented.
}

namespace ensemble create
} ;# namespace ends
